/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 * Copyright (C) 2005-6 Aidan Thornton <makomk@lycos.co.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */


#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

#include "../../src/initng_global.h"
#include "../../src/initng_active_state.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_process_db.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_handler.h"
#include "../../src/main.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_load_module.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_error.h"
#include "../../src/initng_plugin.h"
#include "../../src/initng_static_states.h"
#include "../../src/initng_control_command.h"

#include "../../initng-paths.h"
#include "initng_ngcs.h"
#include "ngcs_marshal.h"
#include "initng_ngcs_cmds.h"

typedef struct ngcs_watch_s
{
    ngcs_channel *chan;
    char *name;
    int flags;
    struct list_head list;
} ngcs_watch;

void register_ngcs_cmds(void);
void unregister_ngcs_cmds(void);
static void ngcs_cmd_halt(ngcs_request * req);
static void ngcs_cmd_reboot(ngcs_request * req);
static int service_status_watch(active_db_h * service);
static void ngcs_cmd_watch(ngcs_request * req);
static void ngcs_free_watch(ngcs_channel * chan);
static int service_output_watch(active_db_h * service, process_h * x, char *buffer_pos);
static ngcs_watch* ngcs_add_watch(ngcs_conn* conn, char* svcname, int flags);
static void ngcs_cmd_start(ngcs_request * req);
static int ngcs_watch_initial(ngcs_watch *watch);
static void ngcs_cmd_stop(ngcs_request * req);
static void ngcs_cmd_hot_reload(ngcs_request * req);

ngcs_cmd ngcs_halt_cmd = {
    "halt",
    ngcs_cmd_halt,
    {0, 0}
};

ngcs_cmd ngcs_reboot_cmd = {
    "reboot",
    ngcs_cmd_reboot,
    {0, 0}
};

ngcs_cmd ngcs_start_cmd = {
    "start",
    ngcs_cmd_start,
    {0, 0}
};


ngcs_cmd ngcs_stop_cmd = {
    "stop",
    ngcs_cmd_stop,
    {0, 0}
};

ngcs_cmd ngcs_watch_cmd = {
    "watch",
    ngcs_cmd_watch,
    {0, 0}
};

ngcs_cmd ngcs_hot_reload_cmd = {
    "hot_reload",
    ngcs_cmd_hot_reload,
    {0, 0}
};

ngcs_watch watches;

static int service_status_watch(active_db_h * service)
{
    ngcs_watch *watch, *nextwatch;
    int len = 0;
    char *buf = NULL;

    len = ngcs_marshal_active_db_h(service, NULL);
    list_for_each_entry_prev_safe(watch, nextwatch, &watches.list, list)
    {
        if (watch->name == NULL || strcmp(watch->name, service->name) == 0)
        {
            if (!buf)
            {
                buf = i_calloc(1, len);
                len = ngcs_marshal_active_db_h(service, buf);
                if (len < 0)
                {
                    F_("ngcs_marshal_active_db_h() failed!\n");
                    free(buf);
                    return TRUE;
                }
            }
	    if(ngcs_channel_send(watch->chan, NGCS_TYPE_STRUCT, len, buf))
	    {
		free(buf); return TRUE;
	    };
       }
    }
    if (buf)
        free(buf);
    return TRUE;
}

static int ngcs_watch_initial(ngcs_watch *watch)
{
    if(watch->flags & NGCS_CURRENT_STATUS)
    {
	active_db_h* current;
	current = NULL;
	while_active_db(current)
	{
	    if(watch->name == NULL || strcmp(watch->name, current->name) == 0)
	    {
		int len = 0; char *buf = NULL;
		len = ngcs_marshal_active_db_h(current, NULL);
		buf = i_calloc(1, len);
		len = ngcs_marshal_active_db_h(current, buf);
		if (len < 0)
                {
                    F_("ngcs_marshal_active_db_h() failed!\n");
                    free(buf);
                    return 1;
                }
		if(ngcs_channel_send(watch->chan, NGCS_TYPE_STRUCT, len, buf))
		{
		    free(buf); return 1;
		};
		free(buf);
	    }
	}
    }
	
    if(ngcs_channel_send(watch->chan, NGCS_TYPE_NULL, 0, NULL))
	return 1;

    return 0;
}

static int service_output_watch(active_db_h * service, process_h * x, char *buffer_pos)
{
    ngcs_watch *watch, *nextwatch;
    ngcs_data dat[2];
    int len = 0;
    char *buf = NULL;;

    dat[0].type = NGCS_TYPE_STRING;
    dat[0].len = -1;
    dat[0].d.s = service->name;
    dat[1].type = NGCS_TYPE_STRING;
    dat[1].len = -1;
    dat[1].d.s = buffer_pos;
    list_for_each_entry_prev_safe(watch, nextwatch, &watches.list, list)
    {
        if (watch->name == NULL || strcmp(watch->name, service->name) == 0)
        {
            if (!buf)
            {
                len = ngcs_pack(dat, 2, NULL);
                assert(len >= 0);
                buf = i_calloc(1, len);
                len = ngcs_pack(dat, 2, buf);
                assert(len >= 0);
            }
            if(ngcs_channel_send(watch->chan, NGCS_TYPE_STRUCT, len, buf))
	    {
		free(buf); return FALSE;
	    };
        }
    }
    if (buf)
        free(buf);
    return FALSE;
}

static void ngcs_cmd_stop(ngcs_request * req)
{
    int i = 0; ngcs_watch* watch;
    active_db_h *serv = NULL; char* svcname = NULL;

    if(req->type != NGCS_TYPE_STRING || req->data == NULL || req->len <= 0)
    {
	F_("Bad call to ngcs command 'stop'\n");
	ngcs_send_response(req, NGCS_TYPE_STRING, 8, (char*)"BAD_CALL");
	return;
    }
    
    svcname = i_calloc(1, req->len+1);
    memcpy(svcname, req->data, req->len);

    serv = initng_active_db_find_in_name(svcname);
    if (!serv)
    {
	ngcs_send_response(req, NGCS_TYPE_STRING, 9, (char*)"NOT_FOUND");
	free(svcname); return;
    }

    watch = ngcs_add_watch(req->conn, serv->name, NGCS_WATCH_STATUS | NGCS_WATCH_OUTPUT | NGCS_CURRENT_STATUS);
    if(watch) i = watch->chan->id;
    ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
    ngcs_watch_initial(watch);
    initng_handler_stop_service(serv);
    free(svcname); return;
}


static void ngcs_cmd_start(ngcs_request * req)
{
    int i = 0; ngcs_watch* watch;
    active_db_h *serv = NULL; char* svcname = NULL;
    if(req->type != NGCS_TYPE_STRING || req->data == NULL || req->len <= 0)
    {
	F_("Bad call to ngcs command 'start'\n");
	ngcs_send_response(req, NGCS_TYPE_STRING, 8, (char*)"BAD_CALL");
	return;
    }
    
    svcname = i_calloc(1, req->len+1);
    memcpy(svcname, req->data, req->len);

    serv = initng_active_db_find_in_name(svcname);
    if (serv)
    {
	watch = ngcs_add_watch(req->conn, serv->name, NGCS_WATCH_STATUS | NGCS_WATCH_OUTPUT | NGCS_CURRENT_STATUS);
	if(watch) i = watch->chan->id;
	ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
	ngcs_watch_initial(watch);
	if (!IS_UP(serv))
            initng_handler_start_service(serv);
	free(svcname); return;
    }
    
    serv = initng_handler_start_new_service_named(svcname);
    if (!serv)
    {
	ngcs_send_response(req, NGCS_TYPE_STRING, 9, (char*)"NOT_FOUND");
	free(svcname); return;
    }

    watch = ngcs_add_watch(req->conn, serv->name, NGCS_WATCH_STATUS | NGCS_WATCH_OUTPUT);
    if(watch) i = watch->chan->id;
    ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
    ngcs_watch_initial(watch); free(svcname); return;
}

ngcs_watch* ngcs_add_watch(ngcs_conn* conn, char* svcname, int flags) 
{
    assert(conn);
    ngcs_channel *chan;
    ngcs_watch *watch;
    watch = i_calloc(1, sizeof(ngcs_watch));

    chan = ngcs_open_channel(conn, NULL, ngcs_free_watch);
    if (!chan)
    {
	F_("ngcs_open_channel failed!\n");
        free(watch); return 0;
    }
    if(svcname)
	watch->name = i_strdup(svcname);
    else watch->name = NULL;
    watch->flags = flags;
    watch->chan = chan;
    watch->list.prev = 0;
    watch->list.next = 0;
    list_add(&watch->list, &watches.list);
    chan->user_data = watch;
    return watch;
}

static void ngcs_cmd_watch(ngcs_request * req)
{
    int i = 0, cnt;
    ngcs_data* data; ngcs_watch *watch;
    char* name = NULL; int flags;

    if(req->type == NGCS_TYPE_STRING)       
    {
	W_("Use of watch(STRING) depreciated - please update client code\n");
       if(req->len <= 0 || !req->data) 
       {
	   name = NULL;
       }
       else
       {
	  name = i_calloc(1, req->len + 1);
	  memcpy(name, req->data, req->len);
       }
       flags = NGCS_WATCH_STATUS | NGCS_WATCH_OUTPUT;
    }
    else if(req->type == NGCS_TYPE_STRUCT)
    {
	cnt = ngcs_unpack(req->data, req->len, &data);
	if(cnt < 0)
	{
	    F_("Bad watch command\n"); 
	    ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
	    return;
	}
	if(cnt != 2 || data[1].type != NGCS_TYPE_INT)
	{
	    F_("Bad watch command\n"); free(data); 
	    ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
	    return;
	}
	flags = data[1].d.i;
	if(data[0].type == NGCS_TYPE_NULL || (data[0].type = NGCS_TYPE_STRING &&
					     data[0].len <= 0))
	{
	    name = NULL;
	}
	else if(data[0].type == NGCS_TYPE_STRING)
	{
	    name = i_strdup(data[0].d.s);
	}
    }
    else if(req->type == NGCS_TYPE_NULL)
    {
	W_("Use of watch(NULL) depreciated - please update client code\n");
        name = NULL;
	flags = NGCS_WATCH_STATUS | NGCS_WATCH_OUTPUT;
    }
    else
    {
	F_("Bad watch command\n");
	ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
	return;
    }
    watch = ngcs_add_watch(req->conn, name, flags);
    if(watch) i = watch->chan->id;
    ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
    if(name) free(name);
    ngcs_watch_initial(watch);
}

static void ngcs_free_watch(ngcs_channel * chan)
{
    ngcs_watch *watch = chan->user_data;

    if (!watch)
        return;
    list_del(&watch->list);
    if (watch->name)
        free(watch->name);
    free(watch);
    chan->user_data = 0;
}

void register_ngcs_cmds(void)
{
    ngcs_reg_cmd(&ngcs_halt_cmd);
    ngcs_reg_cmd(&ngcs_start_cmd);
    ngcs_reg_cmd(&ngcs_stop_cmd);
    ngcs_reg_cmd(&ngcs_reboot_cmd);
    ngcs_reg_cmd(&ngcs_watch_cmd);
    ngcs_reg_cmd(&ngcs_hot_reload_cmd);
    initng_plugin_hook_add(&g.ASTATUS_CHANGE, 50, &service_status_watch);
    initng_plugin_hook_add(&g.PIPEWATCHERS, 50, &service_output_watch);
    INIT_LIST_HEAD(&watches.list);
}

void unregister_ngcs_cmds(void)
{
    initng_plugin_hook_del(&g.ASTATUS_CHANGE, &service_status_watch);
    initng_plugin_hook_del(&g.PIPEWATCHERS, &service_output_watch);
    ngcs_unreg_cmd(&ngcs_halt_cmd);
    ngcs_unreg_cmd(&ngcs_start_cmd);
    ngcs_unreg_cmd(&ngcs_stop_cmd);
    ngcs_unreg_cmd(&ngcs_reboot_cmd);
    ngcs_unreg_cmd(&ngcs_watch_cmd);
    ngcs_unreg_cmd(&ngcs_hot_reload_cmd);
    /* TODO - need to free watches */
}

static void ngcs_cmd_halt(ngcs_request * req)
{
    int i = 0;

    g.when_out = THEN_HALT;
    initng_handler_stop_all();
    ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
}

static void ngcs_cmd_reboot(ngcs_request * req)
{
    int i = 0;

    g.when_out = THEN_REBOOT;
    initng_handler_stop_all();
    ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
}

static void ngcs_cmd_hot_reload(ngcs_request * req)
{
    int i = 0;

    int retval;
    char *new_argv[4];

    retval = initng_plugin_callers_dump_state();

    if (retval == TRUE)
    {
        ngcs_send_response(req, NGCS_TYPE_INT, sizeof(int), (char *) &i);
        D_("exec()int initng\n");
        new_argv[0] = i_strdup("/sbin/initng");
        new_argv[1] = i_strdup("--hot_reload");
        new_argv[2] = NULL;

        execve("/sbin/initng", new_argv, environ);
        F_("Failed to reload initng!\n");
    }
    else if (retval == FALSE)
    {
	F_("No plugin was willing to dump state\n");
	ngcs_send_response(req, NGCS_TYPE_STRING, 13, (char*)"NOT_AVAILABLE");
    }
    else
    {
	F_("dump_state failed!\n");
	ngcs_send_response(req, NGCS_TYPE_STRING, 10, (char*)"DUMP_ERROR");
    }
}
